
package com.loit.component.keygen.leaf;

import com.google.common.base.Preconditions;
import com.google.common.base.Strings;
import com.loit.component.keygen.leaf.core.reg.RegistryCenter;
import com.loit.component.keygen.leaf.core.reg.RegistryCenterConfiguration;
import com.loit.component.keygen.leaf.core.reg.RegistryCenterServiceLoader;
import com.loit.component.keygen.leaf.keygen.LeafKeyGenerator;
import lombok.Getter;
import lombok.Setter;
import lombok.SneakyThrows;

import java.util.Properties;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.SynchronousQueue;

/**
 * Key generator implemented by leaf segment algorithms.
 */
public final class LeafSegmentKeyGenerator implements LeafKeyGenerator {

    protected static final String DEFAULT_NAMESPACE = "leaf_segment";

    protected static final String DEFAULT_STEP = "10000";

    protected static final String DEFAULT_INITIAL_VALUE = "1";

    protected static final String DEFAULT_REGISTRY_CENTER = "zookeeper";

    private static final String SLANTING_BAR = "/";

    private static final String REGULAR_PATTERN = "^((?!/).)*$";

    private final ExecutorService incrementCacheIdExecutor;

    private final SynchronousQueue<Long> cacheIdQueue;

    private RegistryCenter leafRegistryCenter;

    private long id;

    private long step;

    /**
     * 是否第一次加载
     */
    private Boolean isInit = true;

    @Getter
    @Setter
    private Properties properties = new Properties();

    public LeafSegmentKeyGenerator() {
        incrementCacheIdExecutor = Executors.newSingleThreadExecutor();
        cacheIdQueue = new SynchronousQueue<>();
    }

    @Override
    public String getType() {
        return "LEAF_SEGMENT";
    }

    @Override
    public synchronized Comparable<?> generateKey() {
        String leafKey = getLeafKey();
        if (isInit) {
            initLeafSegmentKeyGenerator(leafKey);
        } else {
            increaseIdWhenLeafKeyStoredInCenter(leafKey);
        }

        isInit = false;
        return id;
    }

    /**
     * 构建zookeeper注册中心
     */
    public RegistryCenter buidLeafRegistryCenter() {
        RegistryCenterConfiguration leafConfiguration = getRegistryCenterConfiguration();
        leafRegistryCenter = new RegistryCenterServiceLoader().load(leafConfiguration);

        return leafRegistryCenter;
    }

    /**
     * 设置zookeeper注册中心(共用注册中心)
     */
    public void setLeafRegistryCenter(RegistryCenter registryCenter) {
        this.leafRegistryCenter = registryCenter;
    }

    private void initLeafSegmentKeyGenerator(final String leafKey) {
        buidLeafRegistryCenter();

        step = getStep();
        id = initializeId(leafKey);
        initializeLeafKeyInCenter(leafKey, id, step);
        initializeCacheIdAsynchronous(id, step);
    }

    private void increaseIdWhenLeafKeyStoredInCenter(final String leafKey) {
        if ((id % step) == 0) {
            id = tryTakeCacheId();
            incrementCacheIdAsynchronous(leafKey, step - (id % step));
        }
        ++id;
    }

    private long initializeId(final String leafKey) {
        if (leafRegistryCenter.isExisted(leafKey)) {
            return Long.parseLong(leafRegistryCenter.getDirectly(leafKey)) + 1;
        } else {
            return getInitialValue();
        }
    }

    private void initializeLeafKeyInCenter(final String leafKey, final long id, final long step) {
        leafRegistryCenter.initLock(leafKey + "_lock");
        while (!leafRegistryCenter.tryLock()) {
            continue;
        }
        leafRegistryCenter.persist(leafKey, String.valueOf(id + step - id % step));
        leafRegistryCenter.tryRelease();
    }

    private void initializeCacheIdAsynchronous(final long id, final long step) {
        incrementCacheIdExecutor.execute(new Runnable() {

            @Override
            public void run() {
                tryPutCacheId(id + step - id % step);
            }
        });
    }

    private void incrementCacheIdAsynchronous(final String leafKey, final long step) {
        incrementCacheIdExecutor.execute(new Runnable() {

            @Override
            public void run() {
                long newId = incrementCacheId(leafKey, step);
                tryPutCacheId(newId);
            }
        });
    }

    private RegistryCenterConfiguration getRegistryCenterConfiguration() {
        RegistryCenterConfiguration result = new RegistryCenterConfiguration(getRegistryCenterType(), properties);
        result.setNamespace(properties.getProperty("leaf.namespace", DEFAULT_NAMESPACE));
        result.setServerLists(getServerList());
        result.setDigest(getDigest());
        return result;
    }

    @SneakyThrows
    private long incrementCacheId(final String leafKey, final long step) {
        while (!leafRegistryCenter.tryLock()) {
            continue;
        }
        long result = updateCacheIdInCenter(leafKey, step);
        leafRegistryCenter.tryRelease();
        return result;
    }

    private long updateCacheIdInCenter(final String leafKey, final long step) {
        String cacheIdInString = leafRegistryCenter.getDirectly(leafKey);
        if (Strings.isNullOrEmpty(cacheIdInString)) {
            return Long.MIN_VALUE;
        }
        long cacheId = Long.parseLong(cacheIdInString);
        long result = cacheId + step;
        leafRegistryCenter.update(leafKey, String.valueOf(result));
        return result;
    }

    @SneakyThrows
    private void tryPutCacheId(final long id) {
        cacheIdQueue.put(id);
    }

    @SneakyThrows
    private long tryTakeCacheId() {
        return cacheIdQueue.take();
    }

    private long getStep() {
        long result = Long.parseLong(properties.getProperty("leaf.segment.step", DEFAULT_STEP));
        Preconditions.checkArgument(result > 0L && result < Long.MAX_VALUE);
        return result;
    }

    private long getInitialValue() {
        long result = Long.parseLong(properties.getProperty("leaf.segment.id.initial.value", DEFAULT_INITIAL_VALUE));
        Preconditions.checkArgument(result >= 0L && result < Long.MAX_VALUE);
        return result;
    }

    private String getLeafKey() {
        String leafKey = properties.getProperty("leaf.key");
        Preconditions.checkArgument(!Strings.isNullOrEmpty(leafKey));
        Preconditions.checkArgument(leafKey.matches(REGULAR_PATTERN));
        return SLANTING_BAR + leafKey;
    }

    private String getServerList() {
        String result = properties.getProperty("server.list");
        Preconditions.checkArgument(!Strings.isNullOrEmpty(result));
        return result;
    }

    private String getDigest() {
        return properties.getProperty("registry.center.digest");
    }

    private String getRegistryCenterType() {
        return properties.getProperty("registry.center.type", DEFAULT_REGISTRY_CENTER);
    }
}
